@@ -29,12 +29,11 @@
 #include <glib.h>
 #include "gnome-vfs-mime.h"
 #include "gnome-vfs-mime-info-cache.h"
-#include "gnome-vfs-mime-info.h"
 #include "gnome-vfs-mime-monitor.h"
 #include "gnome-vfs-monitor.h"
 #include "gnome-vfs-ops.h"
-#include "gnome-vfs-utils.h"
-#include "xdgmime.h"
+#include "eggdesktopentries.h"
+#include "eggdirfuncs.h"
 
 typedef struct {
 	char *path;
@@ -49,44 +48,38 @@ typedef struct {
 
 typedef struct {
 	GList *dirs;               /* mimeinfo.cache and defaults.list */
-	GHashTable *global_defaults_cache; /* global results of defaults.list lookup and validation */
+	GList *defaults_list_dirs; /* defaults.list only dirs */
         time_t last_stat_time;
 	guint should_ping_mime_monitor : 1;
 } GnomeVFSMimeInfoCache;
 
 extern void _gnome_vfs_mime_monitor_emit_data_changed (GnomeVFSMIMEMonitor *monitor); 
-extern void _gnome_vfs_mime_info_cache_init (void);
 
+static void free_mime_info_cache_map_list (GList *list);
 static void gnome_vfs_mime_info_cache_dir_init (GnomeVFSMimeInfoCacheDir *dir);
 static void gnome_vfs_mime_info_cache_dir_init_defaults_list (GnomeVFSMimeInfoCacheDir *dir);
 static GnomeVFSMimeInfoCacheDir *gnome_vfs_mime_info_cache_dir_new (const char *path);
 static void gnome_vfs_mime_info_cache_dir_free (GnomeVFSMimeInfoCacheDir *dir);
 static char **gnome_vfs_mime_info_cache_get_search_path (void);
-
+static char **gnome_vfs_mime_info_cache_get_defaults_search_path (void);
+;
 static gboolean gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (GnomeVFSMimeInfoCacheDir *dir,
 								      const char *desktop_entry);
 static void gnome_vfs_mime_info_cache_dir_add_desktop_entries (GnomeVFSMimeInfoCacheDir *dir,
 							       const char *mime_type,
 							       char **new_desktop_file_ids);
+static void gnome_vfs_mime_info_cache_init (void);
 static GnomeVFSMimeInfoCache *gnome_vfs_mime_info_cache_new (void);
 static void gnome_vfs_mime_info_cache_free (GnomeVFSMimeInfoCache *cache);
 
 static GnomeVFSMimeInfoCache *mime_info_cache = NULL;
 G_LOCK_DEFINE_STATIC (mime_info_cache);
 
-
 static void
-destroy_info_cache_value (gpointer key, GList *value, gpointer data)
+free_mime_info_cache_map_list (GList *list)
 {
-	gnome_vfs_list_deep_free (value);
-}
-
-static void
-destroy_info_cache_map (GHashTable *info_cache_map)
-{
-	g_hash_table_foreach (info_cache_map, (GHFunc)destroy_info_cache_value,
-			      NULL);
-	g_hash_table_destroy (info_cache_map);
+	g_list_foreach (list, (GFunc) g_free, NULL);
+	g_list_free (list);
 }
 
 static gboolean
@@ -110,33 +103,19 @@ gnome_vfs_mime_info_cache_dir_out_of_date (GnomeVFSMimeInfoCacheDir *dir, const
 	return FALSE;
 }
 
-/* Call with lock held */
-static gboolean remove_all (gpointer  key,
-			    gpointer  value,
-			    gpointer  user_data)
-{
-	return TRUE;
-}
-
-
-static void
-gnome_vfs_mime_info_cache_blow_global_cache (void)
-{
-	g_hash_table_foreach_remove (mime_info_cache->global_defaults_cache,
-				     remove_all, NULL);
-}
-
 static void
 gnome_vfs_mime_info_cache_dir_init (GnomeVFSMimeInfoCacheDir *dir)
 {
+	EggDesktopEntries *entries;
 	GError *load_error;
-	GKeyFile *key_file;
-	gchar *filename, **mime_types;
+	gchar *filename, **desktop_file_ids, **mime_types;
 	int i;
+	static gchar *allowed_start_groups[] = { "MIME Cache", NULL };
 	struct stat buf;
 
 	load_error = NULL;
 	mime_types = NULL;
+	entries = NULL;
 
 	if (dir->mime_info_cache_map != NULL &&
 	    dir->cache_monitor_handle == NULL &&
@@ -145,14 +124,12 @@ gnome_vfs_mime_info_cache_dir_init (GnomeVFSMimeInfoCacheDir *dir)
 		return;
 
 	if (dir->mime_info_cache_map != NULL) {
-		destroy_info_cache_map (dir->mime_info_cache_map);
+		g_hash_table_destroy (dir->mime_info_cache_map);
 	}
 
 	dir->mime_info_cache_map = g_hash_table_new_full (g_str_hash, g_str_equal,
 							  (GDestroyNotify) g_free,
-							  NULL);
-
-	key_file = g_key_file_new ();
+							  (GDestroyNotify) free_mime_info_cache_map_list);
 
 	filename = g_build_filename (dir->path, "mimeinfo.cache", NULL);
 
@@ -164,27 +141,31 @@ gnome_vfs_mime_info_cache_dir_init (GnomeVFSMimeInfoCacheDir *dir)
 
 	dir->mime_info_cache_timestamp = buf.st_mtime;
 
-	g_key_file_load_from_file (key_file, filename, G_KEY_FILE_NONE, &load_error);
-
+	entries =
+		egg_desktop_entries_new_from_file (allowed_start_groups,
+						   EGG_DESKTOP_ENTRIES_GENERATE_LOOKUP_MAP |
+						   EGG_DESKTOP_ENTRIES_DISCARD_COMMENTS |
+						   EGG_DESKTOP_ENTRIES_DISCARD_TRANSLATIONS,
+						   filename,
+						   &load_error);
 	g_free (filename);
 	filename = NULL;
 
 	if (load_error != NULL)
 		goto error;
 
-	mime_types = g_key_file_get_keys (key_file, "MIME Cache",
-					  NULL, &load_error);
+	mime_types = egg_desktop_entries_get_keys (entries, "MIME Cache",
+						   NULL, &load_error);
 
 	if (load_error != NULL)
 		goto error;
 
 	for (i = 0; mime_types[i] != NULL; i++) {
-		gchar **desktop_file_ids;
-		desktop_file_ids = g_key_file_get_string_list (key_file,
-							       "MIME Cache",
-							       mime_types[i],
-							       NULL,
-							       &load_error);
+		desktop_file_ids = egg_desktop_entries_get_string_list (entries,
+									"MIME Cache",
+									mime_types[i],
+									NULL,
+									&load_error);
 
 		if (load_error != NULL) {
 			g_error_free (load_error);
@@ -200,14 +181,15 @@ gnome_vfs_mime_info_cache_dir_init (GnomeVFSMimeInfoCacheDir *dir)
 	}
 
 	g_strfreev (mime_types);
-	g_key_file_free (key_file);
+	egg_desktop_entries_free (entries);
 
 	return;
 error:
 	if (filename)
 		g_free (filename);
 
-	g_key_file_free (key_file);
+	if (entries != NULL)
+		egg_desktop_entries_free (entries);
 
 	if (mime_types != NULL)
 		g_strfreev (mime_types);
@@ -219,15 +201,16 @@ error:
 static void
 gnome_vfs_mime_info_cache_dir_init_defaults_list (GnomeVFSMimeInfoCacheDir *dir)
 {
-	GKeyFile *key_file;
+	EggDesktopEntries *entries;
 	GError *load_error;
-	gchar *filename, **mime_types;
-	char **desktop_file_ids;
+	gchar *filename, *desktop_file_id, **mime_types;
 	int i;
 	struct stat buf;
+	static gchar *allowed_start_groups[] = { "MIME Cache", NULL };
 
 	load_error = NULL;
 	mime_types = NULL;
+	entries = NULL;
 
 	if (dir->defaults_list_map != NULL &&
 	    dir->defaults_monitor_handle == NULL &&
@@ -240,9 +223,7 @@ gnome_vfs_mime_info_cache_dir_init_defaults_list (GnomeVFSMimeInfoCacheDir *dir)
 	}
 
 	dir->defaults_list_map = g_hash_table_new_full (g_str_hash, g_str_equal,
-							g_free, (GDestroyNotify)g_strfreev);
-
-	key_file = g_key_file_new ();
+							g_free, g_free);
 
 	filename = g_build_filename (dir->path, "defaults.list", NULL);
 
@@ -254,46 +235,78 @@ gnome_vfs_mime_info_cache_dir_init_defaults_list (GnomeVFSMimeInfoCacheDir *dir)
 
 	dir->defaults_list_timestamp = buf.st_mtime;
 
-	g_key_file_load_from_file (key_file, filename, G_KEY_FILE_NONE, &load_error);
-
+	entries =
+		egg_desktop_entries_new_from_file (allowed_start_groups,
+						   EGG_DESKTOP_ENTRIES_GENERATE_LOOKUP_MAP |
+						   EGG_DESKTOP_ENTRIES_DISCARD_COMMENTS |
+						   EGG_DESKTOP_ENTRIES_DISCARD_TRANSLATIONS,
+						   filename,
+						   &load_error);
 	g_free (filename);
 	filename = NULL;
 
 	if (load_error != NULL)
 		goto error;
 
-	mime_types = g_key_file_get_keys (key_file, "Default Applications",
-					  NULL, &load_error);
+	mime_types = egg_desktop_entries_get_keys (entries, "MIME Cache",
+						   NULL, &load_error);
 
 	if (load_error != NULL)
 		goto error;
 
 	for (i = 0; mime_types[i] != NULL; i++) {
-		desktop_file_ids = g_key_file_get_string_list (key_file,
-							       "Default Applications",
-							       mime_types[i],
-							       NULL,
-							       &load_error);
+		GList *tmp;
+
+		desktop_file_id = egg_desktop_entries_get_string (entries,
+								  "MIME Cache",
+								  mime_types[i],
+								  &load_error);
 		if (load_error != NULL) {
 			g_error_free (load_error);
 			load_error = NULL;
 			continue;
 		}
 
-		g_hash_table_replace (dir->defaults_list_map,
-				      g_strdup (mime_types[i]),
-				      desktop_file_ids);
+		/* defaults.list in $XDG_CONFIG_DIRS/mime specify desktop
+		 * files relative to $XDG_DATA_DIRS/applications, so we
+		 * need to iterate through ever dir in $XDG_DATA_DIRS/applications
+		 * looking for a valid desktop file
+		 */
+		if (dir->mime_info_cache_map == NULL) {
+			tmp = mime_info_cache->dirs;
+			while (tmp != NULL) {
+				GnomeVFSMimeInfoCacheDir *app_dir;
+
+				app_dir = (GnomeVFSMimeInfoCacheDir *) tmp->data;
+				if (gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (app_dir,
+											  desktop_file_id)) {
+					g_hash_table_replace (dir->defaults_list_map,
+							      g_strdup (mime_types[i]),
+							      desktop_file_id);
+					break;
+				}
+				tmp = tmp->next;
+			}
+		} else {
+			if (gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (dir,
+										  desktop_file_id)) {
+				g_hash_table_replace (dir->defaults_list_map,
+						      g_strdup (mime_types[i]),
+						      desktop_file_id);
+			}
+		}
 	}
 
 	g_strfreev (mime_types);
-	g_key_file_free (key_file);
+	egg_desktop_entries_free (entries);
 
 	return;
 error:
 	if (filename)
 		g_free (filename);
 
-	g_key_file_free (key_file);
+	if (entries != NULL)
+		egg_desktop_entries_free (entries);
 
 	if (mime_types != NULL)
 		g_strfreev (mime_types);
@@ -302,13 +315,6 @@ error:
 		g_error_free (load_error);
 }
 
-static gboolean
-emit_mime_changed (gpointer data)
-{
-	_gnome_vfs_mime_monitor_emit_data_changed (gnome_vfs_mime_monitor_get ());
-	return FALSE;
-}
-
 static void
 gnome_vfs_mime_info_cache_dir_changed (GnomeVFSMonitorHandle    *handle,
 		                       const gchar              *monitor_uri,
@@ -316,27 +322,8 @@ gnome_vfs_mime_info_cache_dir_changed (GnomeVFSMonitorHandle    *handle,
 		                       GnomeVFSMonitorEventType  event_type,
 	                               GnomeVFSMimeInfoCacheDir *dir)
 {
-	G_LOCK (mime_info_cache);
-	gnome_vfs_mime_info_cache_blow_global_cache ();
 	gnome_vfs_mime_info_cache_dir_init (dir);
-	mime_info_cache->should_ping_mime_monitor = FALSE;
-	G_UNLOCK (mime_info_cache);
-	g_idle_add (emit_mime_changed, NULL);
-}
-
-static void
-gnome_vfs_mime_info_cache_dir_defaults_changed (GnomeVFSMonitorHandle    *handle,
-					 const gchar              *monitor_uri,
-					 const gchar              *info_uri,
-					 GnomeVFSMonitorEventType  event_type,
-					 GnomeVFSMimeInfoCacheDir *dir)
-{
-	G_LOCK (mime_info_cache);
-	gnome_vfs_mime_info_cache_blow_global_cache ();
 	gnome_vfs_mime_info_cache_dir_init_defaults_list (dir);
-	mime_info_cache->should_ping_mime_monitor = FALSE;
-	G_UNLOCK (mime_info_cache);
-	g_idle_add (emit_mime_changed, NULL);
 }
 
 static GnomeVFSMimeInfoCacheDir *
@@ -364,7 +351,7 @@ gnome_vfs_mime_info_cache_dir_new (const char *path)
 			       filename,
 			       GNOME_VFS_MONITOR_FILE,
 			       (GnomeVFSMonitorCallback) 
-			       gnome_vfs_mime_info_cache_dir_defaults_changed,
+			       gnome_vfs_mime_info_cache_dir_changed,
 			       dir);
 	g_free (filename);
 
@@ -377,7 +364,7 @@ gnome_vfs_mime_info_cache_dir_free (GnomeVFSMimeInfoCacheDir *dir)
 	if (dir == NULL)
 		return;
 	if (dir->mime_info_cache_map != NULL) {
-		destroy_info_cache_map (dir->mime_info_cache_map);
+		g_hash_table_destroy (dir->mime_info_cache_map);
 		dir->mime_info_cache_map = NULL;
 
 	}
@@ -404,25 +391,55 @@ static char **
 gnome_vfs_mime_info_cache_get_search_path (void)
 {
 	char **args = NULL;
-	const char * const *data_dirs;
-	const char *user_data_dir;
-	int i, length, j;
+	char **data_dirs;
+	char *user_data_dir;
+	int i, length;
 
-	data_dirs = g_get_system_data_dirs ();
+	data_dirs = egg_get_secondary_data_dirs ();
 
 	for (length = 0; data_dirs[length] != NULL; length++);
 
 	args = g_new (char *, length + 2);
 
-	j = 0;
-	user_data_dir = g_get_user_data_dir ();
-	args[j++] = g_build_filename (user_data_dir, "applications", NULL);
-
-	for (i = 0; i < length; i++) {
-		args[j++] = g_build_filename (data_dirs[i],
-					      "applications", NULL);
+	user_data_dir = egg_get_user_data_dir ();
+	args[length] = g_build_filename (user_data_dir, "applications", NULL);
+	g_free (user_data_dir);
+	
+	i = length - 1;
+	while (i >= 0) {
+		args[length - i - 1] = g_build_filename (data_dirs[i],
+							 "applications", NULL);
+		i--;
 	}
-	args[j++] = NULL;
+	args[length + 1] = NULL;
+
+	g_strfreev (data_dirs);
+
+	return args;
+}
+
+static char **
+gnome_vfs_mime_info_cache_get_defaults_search_path (void)
+{
+	char **args = NULL;
+	char **config_dirs;
+	int i, length;
+
+	config_dirs = egg_get_secondary_configuration_dirs ();
+
+	for (length = 0; config_dirs[length] != NULL; length++);
+
+	args = g_new (char *, length + 1);
+
+	i = length - 1;
+	while (i >= 0) {
+		args[length - i - 1] = g_build_filename (config_dirs[i],
+							 "mime", NULL);
+		i--;
+	}
+	args[length] = NULL;
+
+	g_strfreev (config_dirs);
 
 	return args;
 }
@@ -431,7 +448,7 @@ static gboolean
 gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (GnomeVFSMimeInfoCacheDir *dir,
                                                       const char *desktop_entry)
 {
-	GKeyFile *key_file;
+	EggDesktopEntries *entries;
 	GError *load_error;
 	int i;
 	gboolean can_show_in;
@@ -441,10 +458,12 @@ gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (GnomeVFSMimeInfoCacheDir *
 	can_show_in = TRUE;
 
 	filename = g_build_filename ("applications", desktop_entry, NULL);
-
-	key_file = g_key_file_new ();
-	g_key_file_load_from_data_dirs (key_file, filename, NULL,
-					G_KEY_FILE_NONE, &load_error);
+	entries =
+		egg_desktop_entries_new_from_file (NULL,
+						   EGG_DESKTOP_ENTRIES_DISCARD_COMMENTS |
+						   EGG_DESKTOP_ENTRIES_DISCARD_TRANSLATIONS,
+						   filename,
+						   &load_error);
 	g_free (filename);
 
 	if (load_error != NULL) {
@@ -452,20 +471,21 @@ gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (GnomeVFSMimeInfoCacheDir *
 		return FALSE;
 	}
 
-	if (g_key_file_has_key (key_file, DESKTOP_ENTRY_GROUP,
-				"OnlyShowIn", NULL)) {
+	if (egg_desktop_entries_has_key (entries,
+					 egg_desktop_entries_get_start_group (entries),
+					 "OnlyShowIn")) {
 
 		char **only_show_in_list;
-		only_show_in_list = g_key_file_get_string_list (key_file,
-								DESKTOP_ENTRY_GROUP,
-								"OnlyShowIn",
-								NULL,
-								&load_error);
+		only_show_in_list = egg_desktop_entries_get_string_list (entries,
+									 egg_desktop_entries_get_start_group (entries),
+									 "OnlyShowIn",
+									 NULL,
+									 &load_error);
 
 		if (load_error != NULL) {
 			g_error_free (load_error);
 			g_strfreev (only_show_in_list);
-			g_key_file_free (key_file);
+			egg_desktop_entries_free (entries);
 			return FALSE;
 		}
 
@@ -480,15 +500,15 @@ gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (GnomeVFSMimeInfoCacheDir *
 		g_strfreev (only_show_in_list);
 	}
 
-	if (g_key_file_has_key (key_file,
-				DESKTOP_ENTRY_GROUP,
-				"NotShowIn", NULL)) {
+	if (egg_desktop_entries_has_key (entries,
+					 egg_desktop_entries_get_start_group (entries),
+					 "NotShowIn")) {
 		char **not_show_in_list;
-		not_show_in_list = g_key_file_get_string_list (key_file,
-							       DESKTOP_ENTRY_GROUP,
-							       "NotShowIn",
-							       NULL,
-							       &load_error);
+		not_show_in_list = egg_desktop_entries_get_string_list (entries,
+									egg_desktop_entries_get_start_group (entries),
+									"NotShowIn",
+									NULL,
+									&load_error);
 
 		if (load_error != NULL) {
 			g_error_free (load_error);
@@ -505,28 +525,10 @@ gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (GnomeVFSMimeInfoCacheDir *
 		g_strfreev (not_show_in_list);
 	}
 
-	g_key_file_free (key_file);
+	egg_desktop_entries_free (entries);
 	return can_show_in;
 }
 
-static gboolean
-gnome_vfs_mime_info_desktop_entry_is_valid (const char *desktop_entry)
-{
-	GList *l;
-	
-	for (l = mime_info_cache->dirs; l != NULL; l = l->next) {
-		GnomeVFSMimeInfoCacheDir *app_dir;
-		
-		app_dir = (GnomeVFSMimeInfoCacheDir *) l->data;
-		if (gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (app_dir,
-									  desktop_entry)) {
-			return TRUE;
-		}
-	}
-	return FALSE;
-}
-
-
 static void
 gnome_vfs_mime_info_cache_dir_add_desktop_entries (GnomeVFSMimeInfoCacheDir *dir,
                                                    const char *mime_type,
@@ -539,7 +541,9 @@ gnome_vfs_mime_info_cache_dir_add_desktop_entries (GnomeVFSMimeInfoCacheDir *dir
 						mime_type);
 
 	for (i = 0; new_desktop_file_ids[i] != NULL; i++) {
-		if (!g_list_find (desktop_file_ids, new_desktop_file_ids[i]))
+		if (!g_list_find (desktop_file_ids, new_desktop_file_ids[i]) &&
+		    gnome_vfs_mime_info_cache_dir_desktop_entry_is_valid (dir,
+									  new_desktop_file_ids[i]))
 			desktop_file_ids = g_list_append (desktop_file_ids,
 							  g_strdup (new_desktop_file_ids[i]));
 	}
@@ -571,6 +575,22 @@ gnome_vfs_mime_info_cache_init_dir_lists (void)
 		}
 	}
 	g_strfreev (dirs);
+
+	dirs = gnome_vfs_mime_info_cache_get_defaults_search_path ();
+
+	for (i = 0; dirs[i] != NULL; i++) {
+		GnomeVFSMimeInfoCacheDir *dir;
+
+		dir = gnome_vfs_mime_info_cache_dir_new (dirs[i]);
+
+		if (dir != NULL) {
+			mime_info_cache->defaults_list_dirs =
+				g_list_append (mime_info_cache->defaults_list_dirs, dir);
+			gnome_vfs_mime_info_cache_dir_init_defaults_list (dir);
+		}
+	}
+
+	g_strfreev (dirs);
 }
 
 static void
@@ -583,13 +603,23 @@ gnome_vfs_mime_info_cache_update_dir_lists (void)
 
 		dir = (GnomeVFSMimeInfoCacheDir *) tmp->data;
 
-		if (dir->cache_monitor_handle == NULL) {
-			gnome_vfs_mime_info_cache_blow_global_cache ();
+		if (dir->cache_monitor_handle == NULL) { 
 			gnome_vfs_mime_info_cache_dir_init (dir);
 		}
 		
 		if (dir->defaults_monitor_handle == NULL) {
-			gnome_vfs_mime_info_cache_blow_global_cache ();
+			gnome_vfs_mime_info_cache_dir_init_defaults_list (dir);
+		}
+
+		tmp = tmp->next;
+	}
+
+	tmp = mime_info_cache->defaults_list_dirs;
+	while (tmp != NULL) {
+		GnomeVFSMimeInfoCacheDir *dir;
+
+		dir = (GnomeVFSMimeInfoCacheDir *) tmp->data;
+		if (dir->defaults_monitor_handle == NULL) { 
 			gnome_vfs_mime_info_cache_dir_init_defaults_list (dir);
 		}
 
@@ -597,8 +627,8 @@ gnome_vfs_mime_info_cache_update_dir_lists (void)
 	}
 }
 
-void
-_gnome_vfs_mime_info_cache_init (void)
+static void
+gnome_vfs_mime_info_cache_init (void)
 {
 	G_LOCK (mime_info_cache);
 	if (mime_info_cache == NULL) {
@@ -615,7 +645,7 @@ _gnome_vfs_mime_info_cache_init (void)
 	}
 
 	if (mime_info_cache->should_ping_mime_monitor) {
-		g_idle_add (emit_mime_changed, NULL);
+		_gnome_vfs_mime_monitor_emit_data_changed (gnome_vfs_mime_monitor_get ());
 		mime_info_cache->should_ping_mime_monitor = FALSE;
 	}
 	G_UNLOCK (mime_info_cache);
@@ -629,9 +659,6 @@ gnome_vfs_mime_info_cache_new (void)
 
 	cache = g_new0 (GnomeVFSMimeInfoCache, 1);
 
-	cache->global_defaults_cache = g_hash_table_new_full (g_str_hash, g_str_equal,
-							      (GDestroyNotify) g_free,
-							      (GDestroyNotify) g_free);
 	return cache;
 }
 
@@ -645,8 +672,10 @@ gnome_vfs_mime_info_cache_free (GnomeVFSMimeInfoCache *cache)
 			(GFunc) gnome_vfs_mime_info_cache_dir_free,
 			NULL);
 	g_list_free (cache->dirs);
-	g_hash_table_destroy (cache->global_defaults_cache);
-	g_free (cache);
+	g_list_foreach (cache->defaults_list_dirs,
+			(GFunc) gnome_vfs_mime_info_cache_dir_free,
+			NULL);
+	g_list_free (cache->defaults_list_dirs);
 }
 
 void
@@ -663,156 +692,88 @@ gnome_vfs_mime_info_cache_reload (const char *dir)
 	}
 }
 
-static GList *
-get_all_parent_types (const char *mime_type)
-{
-	const char *umime;
-	const char **parents;
-	GList *l = NULL;
-	int i;
-
-	umime = xdg_mime_unalias_mime_type (mime_type);
-	l = g_list_prepend (l, g_strdup (umime));
-
-	parents = xdg_mime_get_mime_parents (mime_type);
-
-	for (i = 0; parents && parents[i] != NULL; i++) {
-		l = g_list_prepend (l, g_strdup (parents[i]));
-	}
-
-	return g_list_reverse (l);
-}
-
 GList *
 gnome_vfs_mime_get_all_desktop_entries (const char *mime_type)
 {
 	GList *desktop_entries, *list, *dir_list, *tmp;
-	GList *mime_types, *m_list;
-	GnomeVFSMimeInfoCacheDir *dir;
-	char *type;
 
-	_gnome_vfs_mime_info_cache_init ();
+	gnome_vfs_mime_info_cache_init ();
 
 	G_LOCK (mime_info_cache);
-	mime_types = get_all_parent_types (mime_type);
-
+	dir_list = mime_info_cache->dirs;
 	desktop_entries = NULL;
-	for (m_list = mime_types; m_list != NULL; m_list = m_list->next) {
-		type = m_list->data;
-		
-		for (dir_list = mime_info_cache->dirs;
-		     dir_list != NULL;
-		     dir_list = dir_list->next) {
-			dir = (GnomeVFSMimeInfoCacheDir *) dir_list->data;
-			
-			list = g_hash_table_lookup (dir->mime_info_cache_map, type);
-			
-			for (tmp = list; tmp != NULL; tmp = tmp->next) {
-				/* Add if not already in list, and valid */
-				/* TODO: Should we cache here to avoid desktop file validation? */
-				if (!g_list_find_custom (desktop_entries, tmp->data,
-							 (GCompareFunc) strcmp) &&
-				    gnome_vfs_mime_info_desktop_entry_is_valid (tmp->data)) {
-					desktop_entries = g_list_prepend (desktop_entries,
-									  g_strdup (tmp->data));
-				}
+	while (dir_list != NULL) {
+		GnomeVFSMimeInfoCacheDir *dir;
+
+		dir = (GnomeVFSMimeInfoCacheDir *) dir_list->data;
+
+		list = g_hash_table_lookup (dir->mime_info_cache_map, mime_type);
+
+		tmp = list;
+		while (tmp != NULL) {
+			if (!g_list_find_custom (desktop_entries, tmp->data,
+						 (GCompareFunc) strcmp)) {
+				desktop_entries = g_list_prepend (desktop_entries,
+								  g_strdup (tmp->data));
 			}
+			tmp = tmp->next;
 		}
+		dir_list = dir_list->next;
 	}
-
 	G_UNLOCK (mime_info_cache);
 
-	g_list_foreach (mime_types, (GFunc)g_free, NULL);
-	g_list_free (mime_types);
-
 	desktop_entries = g_list_reverse (desktop_entries);
 
 	return desktop_entries;
 }
 
-static gchar *
-get_default_desktop_entry (const char *mime_type)
-{
-	gchar *desktop_entry;
-	char **desktop_entries;
-	GList *dir_list, *l;
-	GList *folder_handlers;
-	GnomeVFSMimeInfoCacheDir *dir;
-	int i;
-
-	desktop_entry = g_hash_table_lookup (mime_info_cache->global_defaults_cache,
-					     mime_type);
-	if (desktop_entry) {
-		return g_strdup (desktop_entry);
-	}
-
-	if (g_str_has_prefix (mime_type, "x-directory/")) {
-		/* We special case folder handler defaults so that
-		 * we always can get a working handler for them
-		 * even though we don't control the defaults.list,
-		 * or the list of availible directory handlers.
-		 */
-
-		for (dir_list = mime_info_cache->dirs; dir_list != NULL; dir_list = dir_list->next) {
-			dir = dir_list->data;
-		
-			folder_handlers = g_hash_table_lookup (dir->mime_info_cache_map, "x-directory/gnome-default-handler");
-			for (l = folder_handlers; l != NULL; l = l->next) {
-				desktop_entry = l->data;
-			
-				if (desktop_entry != NULL &&
-				    gnome_vfs_mime_info_desktop_entry_is_valid (desktop_entry)) {
-					g_hash_table_insert (mime_info_cache->global_defaults_cache,
-							     g_strdup (mime_type), g_strdup (desktop_entry));
-					return g_strdup (desktop_entry);
-				}
-			}
-		}
-	}
-	
-	desktop_entry = NULL;
-	for (dir_list = mime_info_cache->dirs; dir_list != NULL; dir_list = dir_list->next) {
-		dir = dir_list->data;
-		desktop_entries = g_hash_table_lookup (dir->defaults_list_map,
-						       mime_type);
-		for (i = 0; desktop_entries != NULL && desktop_entries[i] != NULL; i++) {
-			desktop_entry = desktop_entries[i];
-			if (desktop_entry != NULL &&
-			    gnome_vfs_mime_info_desktop_entry_is_valid (desktop_entry)) {
-				g_hash_table_insert (mime_info_cache->global_defaults_cache,
-						     g_strdup (mime_type), g_strdup (desktop_entry));
-				return g_strdup (desktop_entry);
-			}
-		}
-	}
-
-	return NULL;
-}
-
 gchar *
 gnome_vfs_mime_get_default_desktop_entry (const char *mime_type)
 {
-	char *desktop_entry = NULL;
-	GList *mime_types, *m_list;
+	gchar *desktop_entry;
+	GList *dir_list;
 
-	_gnome_vfs_mime_info_cache_init ();
+	gnome_vfs_mime_info_cache_init ();
 
 	G_LOCK (mime_info_cache);
+	dir_list = mime_info_cache->defaults_list_dirs;
+	desktop_entry = NULL;
+	while (dir_list != NULL) {
+		GnomeVFSMimeInfoCacheDir *dir;
 
-	mime_types = get_all_parent_types (mime_type);
+		dir = (GnomeVFSMimeInfoCacheDir *) dir_list->data;
 
-	for (m_list = mime_types; m_list != NULL; m_list = m_list->next) {
-		desktop_entry = get_default_desktop_entry (m_list->data);
-		if (desktop_entry) {
+		desktop_entry = g_hash_table_lookup (dir->defaults_list_map,
+						     mime_type);
+
+		if (desktop_entry != NULL)
 			break;
-		}		
-	}
 
+		dir_list = dir_list->next;
+	}
 	G_UNLOCK (mime_info_cache);
 
-	g_list_foreach (mime_types, (GFunc)g_free, NULL);
-	g_list_free (mime_types);
+	if (desktop_entry != NULL)
+		return g_strdup (desktop_entry);
 
-	return desktop_entry;
+	G_LOCK (mime_info_cache);
+	dir_list = mime_info_cache->dirs;
+	desktop_entry = NULL;
+	while (dir_list != NULL) {
+		GnomeVFSMimeInfoCacheDir *dir;
+
+		dir = (GnomeVFSMimeInfoCacheDir *) dir_list->data;
+
+		desktop_entry = g_hash_table_lookup (dir->defaults_list_map,
+						     mime_type);
+
+		if (desktop_entry != NULL)
+			break;
+
+		dir_list = dir_list->next;
+	}
+	G_UNLOCK (mime_info_cache);
+
+	return g_strdup (desktop_entry);
 }
 
