@@ -58,7 +58,8 @@
 #endif
 
 #ifdef HAVE_FAM
-FAMConnection *fam_connection = NULL;
+static FAMConnection *fam_connection = NULL;
+static gint fam_watch_id = 0;
 G_LOCK_DEFINE_STATIC (fam_connection);
 
 typedef struct {
@@ -107,6 +108,7 @@ GET_PATH_MAX (void)
 #define OFF_T off_t
 #endif
 
+
 static gchar *
 get_path_from_uri (GnomeVFSURI const *uri)
 {
@@ -697,13 +699,15 @@ get_stat_info (GnomeVFSFileInfo *file_info,
 				g_free (link_file_path);
 				return gnome_vfs_result_from_errno ();
 			}
-			if (symlink_name[0] != '/') {
+			if ((options & GNOME_VFS_FILE_INFO_FOLLOW_LINKS) &&
+			    symlink_name[0] != '/') {
 				symlink_dir = g_path_get_dirname (link_file_path);
 				newpath = g_build_filename (symlink_dir,
 							    symlink_name, NULL);
 				g_free (symlink_dir);
 				g_free (symlink_name);
-				symlink_name = newpath;
+				symlink_name = gnome_vfs_make_path_name_canonical (newpath);
+				g_free (newpath);
 			}
 			
 			if ((options & GNOME_VFS_FILE_INFO_FOLLOW_LINKS) == 0
@@ -2172,6 +2176,8 @@ fam_do_iter_unlocked (void)
 		if (FAMNextEvent(fam_connection, &ev) != 1) {
 			FAMClose(fam_connection);
 			g_free(fam_connection);
+			g_source_remove (fam_watch_id);
+			fam_watch_id = 0;
 			fam_connection = NULL;
 			return FALSE;
 		}
@@ -2257,13 +2263,12 @@ static gboolean
 monitor_setup (void)
 {
 	GIOChannel *ioc;
-	gint watch_id;
 
 	G_LOCK (fam_connection);
 
 	if (fam_connection == NULL) {
 		fam_connection = g_malloc0(sizeof(FAMConnection));
-		if (FAMOpen2(fam_connection, "test-monitor") != 0) {
+		if (FAMOpen2(fam_connection, "gnome-vfs user") != 0) {
 #ifdef DEBUG_FAM
 			g_print ("FAMOpen failed, FAMErrno=%d\n", FAMErrno);
 #endif
@@ -2273,9 +2278,9 @@ monitor_setup (void)
 			return FALSE;
 		}
 		ioc = g_io_channel_unix_new (FAMCONNECTION_GETFD(fam_connection));
-		watch_id = g_io_add_watch (ioc,
-					   G_IO_IN | G_IO_HUP | G_IO_ERR,
-					   fam_callback, fam_connection);
+		fam_watch_id = g_io_add_watch (ioc,
+					       G_IO_IN | G_IO_HUP | G_IO_ERR,
+					       fam_callback, fam_connection);
 		g_io_channel_unref (ioc);
 	}
 
@@ -2299,11 +2304,15 @@ do_monitor_add (GnomeVFSMethod *method,
 		return GNOME_VFS_ERROR_NOT_SUPPORTED;
 	}
 
+	filename = get_path_from_uri (uri);
+	if (filename == NULL) {
+		return GNOME_VFS_ERROR_INVALID_URI;
+	}
+	
 	handle = g_new0 (FileMonitorHandle, 1);
 	handle->uri = uri;
 	handle->cancelled = FALSE;
 	gnome_vfs_uri_ref (uri);
-	filename = get_path_from_uri (uri);
 
 	G_LOCK (fam_connection);
 	/* We need to queue up incoming messages to avoid blocking on write
